from __future__ import annotations

import logging
import threading
import time
from collections.abc import Callable

from .models import Account
from .store import Store
from .ilink_client import ILinkClient


class PollerManager:
    def __init__(
        self,
        store: Store,
        ilink_client: ILinkClient,
        logger: logging.Logger,
        on_message: Callable[[Account, dict[str, object]], None] | None = None,
    ) -> None:
        self.store = store
        self.ilink_client = ilink_client
        self.logger = logger
        self.on_message = on_message
        self._lock = threading.Lock()
        self._running: dict[str, threading.Event] = {}

    def start_enabled_accounts(self) -> None:
        for account in self.store.list_accounts():
            if account.enabled and account.token:
                self.start_account(account)

    def start_account(self, account: Account) -> None:
        with self._lock:
            if account.account_id in self._running:
                return
            stop_event = threading.Event()
            self._running[account.account_id] = stop_event
        thread = threading.Thread(target=self._run, args=(account, stop_event), daemon=True)
        thread.start()

    def stop_all(self) -> None:
        with self._lock:
            running = list(self._running.values())
            self._running.clear()
        for stop_event in running:
            stop_event.set()

    def stop_account(self, account_id: str) -> None:
        with self._lock:
            stop_event = self._running.pop(account_id, None)
        if stop_event is not None:
            stop_event.set()

    def lookup_context_token(self, account_id: str, peer_user_id: str) -> str:
        try:
            return self.store.get_peer_context(account_id, peer_user_id).context_token
        except LookupError:
            return ""

    def _run(self, account: Account, stop_event: threading.Event) -> None:
        log = self.logger.getChild(f"poller.{account.account_id}")
        current = account
        backoff = 2.0
        log.info("poller started")
        try:
            while not stop_event.is_set():
                try:
                    payload = self.ilink_client.get_updates(current.base_url, current.token, current.get_updates_buf)
                except Exception as exc:
                    self.store.update_account_poll_state(current.account_id, current.get_updates_buf, "error", str(exc))
                    log.error("getupdates failed: %s", exc)
                    if stop_event.wait(backoff):
                        return
                    backoff = min(backoff * 2, 30.0)
                    continue

                backoff = 2.0
                current.get_updates_buf = str(payload.get("get_updates_buf") or current.get_updates_buf)
                ret = int(payload.get("ret", 0) or 0)
                err_code = int(payload.get("errcode", 0) or 0)
                status = "connected"
                err_text = ""
                if ret != 0 or err_code != 0:
                    status = "warning"
                    err_text = str(payload.get("errmsg") or "getupdates returned non-zero status")
                self.store.update_account_poll_state(current.account_id, current.get_updates_buf, status, err_text)

                for message in list(payload.get("msgs") or []):
                    msg = dict(message)
                    if int(msg.get("message_type", 0) or 0) != 1:
                        continue
                    try:
                        if self.on_message is None:
                            self.store.save_inbound_message(current.account_id, msg, "", "", "")
                        else:
                            self.on_message(current, msg)
                    except Exception as exc:
                        log.error("save inbound event failed: %s", exc)

                try:
                    current = self.store.get_account(current.account_id)
                except LookupError:
                    return

                delay = 0.0 if int(payload.get("longpolling_timeout_ms", 0) or 0) > 0 else 1.0
                if delay > 0 and stop_event.wait(delay):
                    return
        finally:
            with self._lock:
                self._running.pop(account.account_id, None)
            log.info("poller stopped")



from __future__ import annotations

import json
import logging
from dataclasses import is_dataclass
from datetime import datetime
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from threading import Thread
from typing import Any
from urllib import parse

from .models import Account, Event, LogEntry, LoginSession, Settings, VersionInfo
from .qr import generate_qrcode_png
from .service import Service
from .version import current


class APIServer:
    def __init__(self, service: Service, logger: logging.Logger, listen_addr: str) -> None:
        host, port_text = listen_addr.rsplit(":", 1)
        self._server = ThreadingHTTPServer((host, int(port_text)), _make_handler(service, logger))
        self._thread: Thread | None = None

    def start(self) -> None:
        if self._thread is not None:
            return
        self._thread = Thread(target=self._server.serve_forever, daemon=True)
        self._thread.start()

    def shutdown(self) -> None:
        self._server.shutdown()
        self._server.server_close()
        if self._thread is not None:
            self._thread.join(timeout=5)


def _make_handler(service: Service, logger: logging.Logger):
    class Handler(BaseHTTPRequestHandler):
        def do_GET(self) -> None:
            path, _, query_string = self.path.partition("?")
            query = parse.parse_qs(query_string)
            try:
                if path == "/health/live" or path == "/health/ready":
                    self._write_json(200, {"ok": True, "timestamp": datetime.utcnow().isoformat() + "Z", "version": current()})
                    return
                if path == "/api/version":
                    self._write_json(200, current())
                    return
                if path == "/api/accounts/login/status":
                    session_id = query_value(query, "session_id")
                    self._write_json(200, service.get_login_status(session_id))
                    return
                if path == "/api/accounts/login/qr":
                    session_id = query_value(query, "session_id")
                    session = service.get_login_session(session_id)
                    if not session.qr_code_url:
                        raise ValueError("qr code url is empty")
                    payload = generate_qrcode_png(session.qr_code_url)
                    self.send_response(200)
                    self.send_header("Content-Type", "image/png")
                    self.end_headers()
                    self.wfile.write(payload)
                    return
                if path == "/api/accounts":
                    self._write_json(200, {"items": service.list_accounts()})
                    return
                if path == "/api/events":
                    self._write_json(
                        200,
                        {"items": service.list_events(int(query.get("after_id", ["0"])[0]), int(query.get("limit", ["100"])[0]))},
                    )
                    return
                if path == "/api/logs":
                    self._write_json(
                        200,
                        {"items": service.list_logs(int(query.get("after_id", ["0"])[0]), int(query.get("limit", ["100"])[0]))},
                    )
                    return
                if path == "/api/settings":
                    self._write_json(200, service.get_settings())
                    return
                self._write_json(404, {"error": "not found"})
            except LookupError as exc:
                self._write_json(404, {"error": str(exc)})
            except ValueError as exc:
                self._write_json(400, {"error": str(exc)})
            except Exception as exc:
                logger.exception("request failed")
                self._write_json(500, {"error": str(exc)})

        def do_POST(self) -> None:
            path = self.path.split("?", 1)[0]
            try:
                payload = self._read_json()
                if path == "/api/accounts/login/start":
                    self._write_json(200, service.start_login(str(payload.get("base_url") or "")))
                    return
                if path == "/api/settings":
                    settings = Settings.from_dict(payload)
                    if not settings.listen_addr.strip():
                        raise ValueError("listen_addr is required")
                    out = service.update_settings(settings)
                    self._write_json(200, {"settings": out, "restart_needed": True})
                    return
                if path == "/api/messages/send-text":
                    service.send_text(
                        str(payload.get("account_id") or ""),
                        str(payload.get("to_user_id") or ""),
                        str(payload.get("text") or ""),
                        str(payload.get("context_token") or ""),
                    )
                    self._write_json(200, {"ok": True})
                    return
                if path == "/api/messages/send-media":
                    service.send_media(
                        str(payload.get("account_id") or ""),
                        str(payload.get("to_user_id") or ""),
                        str(payload.get("type") or ""),
                        str(payload.get("file_path") or ""),
                        str(payload.get("text") or ""),
                        str(payload.get("context_token") or ""),
                    )
                    self._write_json(200, {"ok": True})
                    return
                self._write_json(404, {"error": "not found"})
            except LookupError as exc:
                self._write_json(404, {"error": str(exc)})
            except ValueError as exc:
                self._write_json(400, {"error": str(exc)})
            except Exception as exc:
                logger.exception("request failed")
                self._write_json(500, {"error": str(exc)})

        def log_message(self, format: str, *args: object) -> None:
            logger.debug("%s - %s", self.address_string(), format % args)

        def _read_json(self) -> dict[str, Any]:
            content_length = int(self.headers.get("Content-Length", "0") or "0")
            if content_length == 0:
                return {}
            raw = self.rfile.read(content_length)
            return json.loads(raw.decode("utf-8"))

        def _write_json(self, status: int, payload: Any) -> None:
            raw = json.dumps(serialize(payload), ensure_ascii=False).encode("utf-8")
            self.send_response(status)
            self.send_header("Content-Type", "application/json; charset=utf-8")
            self.send_header("Content-Length", str(len(raw)))
            self.end_headers()
            self.wfile.write(raw)

    return Handler


def serialize(value: Any) -> Any:
    if isinstance(value, (LoginSession, Account, Event, LogEntry, Settings, VersionInfo)):
        if isinstance(value, (LoginSession, Account)):
            return value.to_dict()
        return value.to_dict()
    if isinstance(value, list):
        return [serialize(item) for item in value]
    if isinstance(value, dict):
        return {key: serialize(item) for key, item in value.items()}
    if is_dataclass(value):
        return serialize(value.__dict__)
    return value


def query_value(query: dict[str, list[str]], key: str) -> str:
    value = query.get(key, [""])[0].strip()
    if not value:
        raise ValueError(f"{key} is required")
    return value



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any


def utc_now_iso() -> str:
    return datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")


def _as_str(value: Any, default: str = "") -> str:
    if value is None:
        return default
    return str(value)


def _as_int(value: Any, default: int = 0) -> int:
    if value in (None, ""):
        return default
    try:
        return int(value)
    except (TypeError, ValueError):
        return default


def _as_bool(value: Any, default: bool = False) -> bool:
    if value is None:
        return default
    if isinstance(value, bool):
        return value
    if isinstance(value, str):
        normalized = value.strip().lower()
        if normalized in {"1", "true", "yes", "on"}:
            return True
        if normalized in {"0", "false", "no", "off"}:
            return False
    return bool(value)


@dataclass(slots=True)
class VersionInfo:
    version: str
    commit: str | None = None
    build_time: str | None = None
    modified: bool = False

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "VersionInfo":
        return cls(
            version=_as_str(data.get("version")),
            commit=_as_str(data.get("commit")) or None,
            build_time=_as_str(data.get("build_time")) or None,
            modified=_as_bool(data.get("modified"), False),
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            "version": self.version,
            "commit": self.commit,
            "build_time": self.build_time,
            "modified": self.modified,
        }


@dataclass(slots=True)
class LoginSession:
    session_id: str
    base_url: str
    qr_code: str = ""
    qr_code_url: str = ""
    status: str = ""
    account_id: str = ""
    ilink_user_id: str = ""
    bot_token: str = ""
    error: str = ""
    started_at: str = ""
    updated_at: str = ""
    completed_at: str = ""

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "LoginSession":
        return cls(
            session_id=_as_str(data.get("session_id")),
            base_url=_as_str(data.get("base_url")),
            qr_code=_as_str(data.get("qr_code")),
            qr_code_url=_as_str(data.get("qr_code_url")),
            status=_as_str(data.get("status")),
            account_id=_as_str(data.get("account_id")),
            ilink_user_id=_as_str(data.get("ilink_user_id")),
            bot_token=_as_str(data.get("bot_token")),
            error=_as_str(data.get("error")),
            started_at=_as_str(data.get("started_at")),
            updated_at=_as_str(data.get("updated_at")),
            completed_at=_as_str(data.get("completed_at")),
        )

    def to_dict(self, include_private: bool = False) -> dict[str, Any]:
        data = {
            "session_id": self.session_id,
            "base_url": self.base_url,
            "qr_code_url": self.qr_code_url,
            "status": self.status,
            "account_id": self.account_id,
            "ilink_user_id": self.ilink_user_id,
            "error": self.error,
            "started_at": self.started_at,
            "updated_at": self.updated_at,
            "completed_at": self.completed_at,
        }
        if include_private:
            data["qr_code"] = self.qr_code
            data["bot_token"] = self.bot_token
        return data


@dataclass(slots=True)
class Account:
    account_id: str
    base_url: str
    token: str = ""
    ilink_user_id: str = ""
    enabled: bool = False
    login_status: str = ""
    last_error: str = ""
    get_updates_buf: str = ""
    last_poll_at: str = ""
    last_inbound_at: str = ""
    created_at: str = ""
    updated_at: str = ""

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "Account":
        return cls(
            account_id=_as_str(data.get("account_id")),
            base_url=_as_str(data.get("base_url")),
            token=_as_str(data.get("token")),
            ilink_user_id=_as_str(data.get("ilink_user_id")),
            enabled=_as_bool(data.get("enabled"), False),
            login_status=_as_str(data.get("login_status")),
            last_error=_as_str(data.get("last_error")),
            get_updates_buf=_as_str(data.get("get_updates_buf")),
            last_poll_at=_as_str(data.get("last_poll_at")),
            last_inbound_at=_as_str(data.get("last_inbound_at")),
            created_at=_as_str(data.get("created_at")),
            updated_at=_as_str(data.get("updated_at")),
        )

    def to_dict(self, include_private: bool = False) -> dict[str, Any]:
        data = {
            "account_id": self.account_id,
            "base_url": self.base_url,
            "ilink_user_id": self.ilink_user_id,
            "enabled": self.enabled,
            "login_status": self.login_status,
            "last_error": self.last_error,
            "last_poll_at": self.last_poll_at,
            "last_inbound_at": self.last_inbound_at,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
        }
        if include_private:
            data["token"] = self.token
            data["get_updates_buf"] = self.get_updates_buf
        return data


@dataclass(slots=True)
class PeerContext:
    account_id: str
    peer_user_id: str
    context_token: str
    updated_at: str

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "PeerContext":
        return cls(
            account_id=_as_str(data.get("account_id")),
            peer_user_id=_as_str(data.get("peer_user_id")),
            context_token=_as_str(data.get("context_token")),
            updated_at=_as_str(data.get("updated_at")),
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            "account_id": self.account_id,
            "peer_user_id": self.peer_user_id,
            "context_token": self.context_token,
            "updated_at": self.updated_at,
        }


@dataclass(slots=True)
class Event:
    id: int
    account_id: str
    direction: str
    event_type: str
    from_user_id: str = ""
    to_user_id: str = ""
    message_id: int = 0
    context_token: str = ""
    body_text: str = ""
    media_path: str = ""
    media_file_name: str = ""
    media_mime_type: str = ""
    raw_json: str = ""
    created_at: str = ""

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "Event":
        return cls(
            id=_as_int(data.get("id")),
            account_id=_as_str(data.get("account_id")),
            direction=_as_str(data.get("direction")),
            event_type=_as_str(data.get("event_type")),
            from_user_id=_as_str(data.get("from_user_id")),
            to_user_id=_as_str(data.get("to_user_id")),
            message_id=_as_int(data.get("message_id")),
            context_token=_as_str(data.get("context_token")),
            body_text=_as_str(data.get("body_text")),
            media_path=_as_str(data.get("media_path")),
            media_file_name=_as_str(data.get("media_file_name")),
            media_mime_type=_as_str(data.get("media_mime_type")),
            raw_json=_as_str(data.get("raw_json")),
            created_at=_as_str(data.get("created_at")),
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "account_id": self.account_id,
            "direction": self.direction,
            "event_type": self.event_type,
            "from_user_id": self.from_user_id,
            "to_user_id": self.to_user_id,
            "message_id": self.message_id,
            "context_token": self.context_token,
            "body_text": self.body_text,
            "media_path": self.media_path,
            "media_file_name": self.media_file_name,
            "media_mime_type": self.media_mime_type,
            "raw_json": self.raw_json,
            "created_at": self.created_at,
        }


@dataclass(slots=True)
class LogEntry:
    id: int
    level: str
    message: str
    source: str
    meta_json: str = ""
    created_at: str = ""

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "LogEntry":
        return cls(
            id=_as_int(data.get("id")),
            level=_as_str(data.get("level")),
            message=_as_str(data.get("message")),
            source=_as_str(data.get("source")),
            meta_json=_as_str(data.get("meta_json")),
            created_at=_as_str(data.get("created_at")),
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "level": self.level,
            "message": self.message,
            "source": self.source,
            "meta_json": self.meta_json,
            "created_at": self.created_at,
        }


@dataclass(slots=True)
class Settings:
    listen_addr: str
    webhook_url: str = ""

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "Settings":
        return cls(
            listen_addr=_as_str(data.get("listen_addr")),
            webhook_url=_as_str(data.get("webhook_url")),
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            "listen_addr": self.listen_addr,
            "webhook_url": self.webhook_url,
        }



from __future__ import annotations

import json
import sqlite3
import threading
from pathlib import Path

from .models import Account, Event, LogEntry, LoginSession, PeerContext, utc_now_iso


class Store:
    def __init__(self, db_path: str) -> None:
        Path(db_path).parent.mkdir(parents=True, exist_ok=True)
        self._lock = threading.RLock()
        self._conn = sqlite3.connect(db_path, check_same_thread=False)
        self._conn.row_factory = sqlite3.Row
        self._migrate()

    def close(self) -> None:
        with self._lock:
            self._conn.close()

    def ping(self) -> None:
        with self._lock:
            self._conn.execute("SELECT 1")

    def create_login_session(self, session: LoginSession) -> None:
        with self._lock, self._conn:
            self._conn.execute(
                """
                INSERT INTO login_sessions (
                  session_id, base_url, qr_code, qr_code_url, status, error, started_at, updated_at
                ) VALUES (?, ?, ?, ?, ?, ?, ?, ?)
                """,
                (
                    session.session_id,
                    session.base_url,
                    session.qr_code,
                    session.qr_code_url,
                    session.status,
                    session.error,
                    session.started_at,
                    session.updated_at,
                ),
            )

    def get_login_session(self, session_id: str) -> LoginSession:
        row = self._fetchone(
            """
            SELECT session_id, base_url, qr_code, qr_code_url, status, account_id, ilink_user_id, bot_token,
                   error, started_at, updated_at, completed_at
            FROM login_sessions WHERE session_id = ?
            """,
            (session_id,),
        )
        if row is None:
            raise LookupError("login session not found")
        return LoginSession.from_dict(dict(row))

    def update_login_session_status(self, session_id: str, status: str, error_text: str) -> None:
        with self._lock, self._conn:
            self._conn.execute(
                """
                UPDATE login_sessions
                SET status = ?, error = ?, updated_at = ?
                WHERE session_id = ?
                """,
                (status, error_text, utc_now_iso(), session_id),
            )

    def complete_login_session(self, session_id: str, status: dict[str, str]) -> None:
        now = utc_now_iso()
        base_url = str(status.get("baseurl") or "")
        with self._lock, self._conn:
            if not base_url:
                row = self._fetchone("SELECT base_url FROM login_sessions WHERE session_id = ?", (session_id,))
                base_url = str(row["base_url"]) if row else ""
            self._conn.execute(
                """
                UPDATE login_sessions
                SET status = ?, account_id = ?, ilink_user_id = ?, bot_token = ?, base_url = ?, updated_at = ?, completed_at = ?
                WHERE session_id = ?
                """,
                (
                    str(status.get("status") or ""),
                    str(status.get("ilink_bot_id") or ""),
                    str(status.get("ilink_user_id") or ""),
                    str(status.get("bot_token") or ""),
                    base_url,
                    now,
                    now,
                    session_id,
                ),
            )
            self._conn.execute(
                """
                INSERT INTO accounts (
                  account_id, base_url, token, ilink_user_id, enabled, login_status, created_at, updated_at
                ) VALUES (?, ?, ?, ?, 1, 'connected', ?, ?)
                ON CONFLICT(account_id) DO UPDATE SET
                  base_url = excluded.base_url,
                  token = excluded.token,
                  ilink_user_id = excluded.ilink_user_id,
                  enabled = 1,
                  login_status = 'connected',
                  last_error = '',
                  updated_at = excluded.updated_at
                """,
                (
                    str(status.get("ilink_bot_id") or ""),
                    base_url,
                    str(status.get("bot_token") or ""),
                    str(status.get("ilink_user_id") or ""),
                    now,
                    now,
                ),
            )

    def list_accounts(self) -> list[Account]:
        rows = self._fetchall(
            """
            SELECT account_id, base_url, token, ilink_user_id, enabled, login_status, last_error,
                   get_updates_buf, last_poll_at, last_inbound_at, created_at, updated_at
            FROM accounts
            ORDER BY created_at ASC
            """
        )
        return [Account.from_dict(dict(row)) for row in rows]

    def get_account(self, account_id: str) -> Account:
        row = self._fetchone(
            """
            SELECT account_id, base_url, token, ilink_user_id, enabled, login_status, last_error,
                   get_updates_buf, last_poll_at, last_inbound_at, created_at, updated_at
            FROM accounts WHERE account_id = ?
            """,
            (account_id,),
        )
        if row is None:
            raise LookupError("account not found")
        return Account.from_dict(dict(row))

    def delete_account(self, account_id: str) -> None:
        with self._lock, self._conn:
            self._conn.execute("DELETE FROM accounts WHERE account_id = ?", (account_id,))
            self._conn.execute("DELETE FROM peer_contexts WHERE account_id = ?", (account_id,))
            self._conn.execute("DELETE FROM login_sessions WHERE account_id = ?", (account_id,))

    def update_account_poll_state(self, account_id: str, get_updates_buf: str, login_status: str, last_error: str) -> None:
        now = utc_now_iso()
        with self._lock, self._conn:
            self._conn.execute(
                """
                UPDATE accounts
                SET get_updates_buf = ?, login_status = ?, last_error = ?, last_poll_at = ?, updated_at = ?
                WHERE account_id = ?
                """,
                (get_updates_buf, login_status, last_error, now, now, account_id),
            )

    def save_inbound_message(
        self,
        account_id: str,
        message: dict[str, object],
        media_path: str,
        media_file_name: str,
        media_mime_type: str,
    ) -> None:
        raw = json.dumps(message, ensure_ascii=False)
        body_text = extract_body_text(message)
        now = utc_now_iso()
        with self._lock, self._conn:
            self._conn.execute(
                """
                INSERT OR IGNORE INTO events (
                  account_id, direction, event_type, from_user_id, to_user_id, message_id, context_token,
                  body_text, media_path, media_file_name, media_mime_type, raw_json, created_at
                ) VALUES (?, 'inbound', ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                """,
                (
                    account_id,
                    detect_event_type(message),
                    str(message.get("from_user_id") or ""),
                    str(message.get("to_user_id") or ""),
                    int(message.get("message_id") or 0),
                    str(message.get("context_token") or ""),
                    body_text,
                    media_path,
                    media_file_name,
                    media_mime_type,
                    raw,
                    now,
                ),
            )
            from_user_id = str(message.get("from_user_id") or "")
            context_token = str(message.get("context_token") or "")
            if from_user_id and context_token:
                self._conn.execute(
                    """
                    INSERT INTO peer_contexts (account_id, peer_user_id, context_token, updated_at)
                    VALUES (?, ?, ?, ?)
                    ON CONFLICT(account_id, peer_user_id) DO UPDATE SET
                      context_token = excluded.context_token,
                      updated_at = excluded.updated_at
                    """,
                    (account_id, from_user_id, context_token, now),
                )
            self._conn.execute(
                """
                UPDATE accounts
                SET last_inbound_at = ?, updated_at = ?, last_error = '', login_status = 'connected'
                WHERE account_id = ?
                """,
                (now, now, account_id),
            )

    def get_peer_context(self, account_id: str, peer_user_id: str) -> PeerContext:
        row = self._fetchone(
            """
            SELECT account_id, peer_user_id, context_token, updated_at
            FROM peer_contexts WHERE account_id = ? AND peer_user_id = ?
            """,
            (account_id, peer_user_id),
        )
        if row is None:
            raise LookupError("peer context not found")
        return PeerContext.from_dict(dict(row))

    def create_outbound_event(
        self,
        account_id: str,
        event_type: str,
        to_user_id: str,
        context_token: str,
        body_text: str,
        media_path: str,
        media_file_name: str,
        media_mime_type: str,
        raw_json: str,
    ) -> None:
        with self._lock, self._conn:
            self._conn.execute(
                """
                INSERT INTO events (
                  account_id, direction, event_type, to_user_id, context_token, body_text, media_path,
                  media_file_name, media_mime_type, raw_json, created_at
                ) VALUES (?, 'outbound', ?, ?, ?, ?, ?, ?, ?, ?, ?)
                """,
                (
                    account_id,
                    event_type,
                    to_user_id,
                    context_token,
                    body_text,
                    media_path,
                    media_file_name,
                    media_mime_type,
                    raw_json,
                    utc_now_iso(),
                ),
            )

    def add_log(self, level: str, message: str, source: str, meta_json: str = "") -> None:
        with self._lock, self._conn:
            self._conn.execute(
                "INSERT INTO logs (level, message, source, meta_json, created_at) VALUES (?, ?, ?, ?, ?)",
                (level, message, source, meta_json, utc_now_iso()),
            )

    def list_logs(self, after_id: int = 0, limit: int = 100) -> list[LogEntry]:
        if limit <= 0 or limit > 500:
            limit = 100
        rows = self._fetchall(
            """
            SELECT id, level, message, source, meta_json, created_at
            FROM logs WHERE id > ? ORDER BY id ASC LIMIT ?
            """,
            (after_id, limit),
        )
        return [LogEntry.from_dict(dict(row)) for row in rows]

    def list_events(self, after_id: int = 0, limit: int = 100) -> list[Event]:
        if limit <= 0 or limit > 500:
            limit = 100
        rows = self._fetchall(
            """
            SELECT id, account_id, direction, event_type, from_user_id, to_user_id, message_id, context_token,
                   body_text, media_path, media_file_name, media_mime_type, raw_json, created_at
            FROM events WHERE id > ? ORDER BY id ASC LIMIT ?
            """,
            (after_id, limit),
        )
        return [Event.from_dict(dict(row)) for row in rows]

    def _migrate(self) -> None:
        statements = [
            "PRAGMA journal_mode=WAL;",
            """
            CREATE TABLE IF NOT EXISTS login_sessions (
                session_id TEXT PRIMARY KEY,
                base_url TEXT NOT NULL,
                qr_code TEXT NOT NULL,
                qr_code_url TEXT NOT NULL,
                status TEXT NOT NULL,
                account_id TEXT NOT NULL DEFAULT '',
                ilink_user_id TEXT NOT NULL DEFAULT '',
                bot_token TEXT NOT NULL DEFAULT '',
                error TEXT NOT NULL DEFAULT '',
                started_at TEXT NOT NULL,
                updated_at TEXT NOT NULL,
                completed_at TEXT NOT NULL DEFAULT ''
            )
            """,
            """
            CREATE TABLE IF NOT EXISTS accounts (
                account_id TEXT PRIMARY KEY,
                base_url TEXT NOT NULL,
                token TEXT NOT NULL DEFAULT '',
                ilink_user_id TEXT NOT NULL DEFAULT '',
                enabled INTEGER NOT NULL DEFAULT 1,
                login_status TEXT NOT NULL DEFAULT 'pending',
                last_error TEXT NOT NULL DEFAULT '',
                get_updates_buf TEXT NOT NULL DEFAULT '',
                last_poll_at TEXT NOT NULL DEFAULT '',
                last_inbound_at TEXT NOT NULL DEFAULT '',
                created_at TEXT NOT NULL,
                updated_at TEXT NOT NULL
            )
            """,
            """
            CREATE TABLE IF NOT EXISTS peer_contexts (
                account_id TEXT NOT NULL,
                peer_user_id TEXT NOT NULL,
                context_token TEXT NOT NULL,
                updated_at TEXT NOT NULL,
                PRIMARY KEY (account_id, peer_user_id)
            )
            """,
            """
            CREATE TABLE IF NOT EXISTS events (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                account_id TEXT NOT NULL,
                direction TEXT NOT NULL,
                event_type TEXT NOT NULL,
                from_user_id TEXT NOT NULL DEFAULT '',
                to_user_id TEXT NOT NULL DEFAULT '',
                message_id INTEGER NOT NULL DEFAULT 0,
                context_token TEXT NOT NULL DEFAULT '',
                body_text TEXT NOT NULL DEFAULT '',
                media_path TEXT NOT NULL DEFAULT '',
                media_file_name TEXT NOT NULL DEFAULT '',
                media_mime_type TEXT NOT NULL DEFAULT '',
                raw_json TEXT NOT NULL,
                created_at TEXT NOT NULL
            )
            """,
            """
            CREATE UNIQUE INDEX IF NOT EXISTS idx_events_account_message_inbound
            ON events(account_id, direction, message_id)
            WHERE direction = 'inbound' AND message_id != 0
            """,
            """
            CREATE TABLE IF NOT EXISTS logs (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                level TEXT NOT NULL,
                message TEXT NOT NULL,
                source TEXT NOT NULL,
                meta_json TEXT NOT NULL DEFAULT '',
                created_at TEXT NOT NULL
            )
            """,
        ]
        with self._lock, self._conn:
            for stmt in statements:
                self._conn.execute(stmt)

    def _fetchone(self, sql: str, params: tuple[object, ...] = ()) -> sqlite3.Row | None:
        with self._lock:
            return self._conn.execute(sql, params).fetchone()

    def _fetchall(self, sql: str, params: tuple[object, ...] = ()) -> list[sqlite3.Row]:
        with self._lock:
            return self._conn.execute(sql, params).fetchall()


def extract_body_text(message: dict[str, object]) -> str:
    for item in list(message.get("item_list") or []):
        entry = dict(item)
        item_type = int(entry.get("type", 0) or 0)
        if item_type == 1:
            text_item = dict(entry.get("text_item") or {})
            return str(text_item.get("text") or "")
        if item_type == 3:
            voice_item = dict(entry.get("voice_item") or {})
            if voice_item.get("text"):
                return str(voice_item["text"])
        if item_type == 2:
            return "[image]"
        if item_type == 4:
            file_item = dict(entry.get("file_item") or {})
            file_name = str(file_item.get("file_name") or "")
            return f"[file] {file_name}".rstrip()
        if item_type == 5:
            return "[video]"
    return ""


def detect_event_type(message: dict[str, object]) -> str:
    for item in list(message.get("item_list") or []):
        item_type = int(dict(item).get("type", 0) or 0)
        if item_type == 1:
            return "text"
        if item_type == 2:
            return "image"
        if item_type == 3:
            return "voice"
        if item_type == 4:
            return "file"
        if item_type == 5:
            return "video"
    return "unknown"



from __future__ import annotations

import base64
import hashlib
import json
import mimetypes
import os
import secrets
from pathlib import Path
from typing import Any
from urllib import error, parse, request


class ILinkClient:
    def __init__(self, channel_version: str, timeout: float = 45.0) -> None:
        self.channel_version = channel_version
        self.timeout = timeout

    def fetch_qrcode(self, base_url: str) -> dict[str, Any]:
        endpoint = f"{base_url.rstrip('/')}/ilink/bot/get_bot_qrcode?bot_type=3"
        req = request.Request(endpoint, method="GET")
        return self._do_json(req, token="", payload=None)

    def fetch_qrcode_status(self, base_url: str, qr_code: str) -> dict[str, Any]:
        endpoint = f"{base_url.rstrip('/')}/ilink/bot/get_qrcode_status?qrcode={parse.quote(qr_code)}"
        req = request.Request(endpoint, method="GET", headers={"iLink-App-ClientVersion": "1"})
        return self._do_json(req, token="", payload=None)

    def get_updates(self, base_url: str, token: str, get_updates_buf: str) -> dict[str, Any]:
        return self._post_json(
            f"{base_url.rstrip('/')}/ilink/bot/getupdates",
            token,
            {
                "get_updates_buf": get_updates_buf,
                "base_info": {"channel_version": self.channel_version},
            },
        )

    def send_text_message(self, base_url: str, token: str, to_user_id: str, text: str, context_token: str) -> None:
        payload = {
            "msg": {
                "from_user_id": "",
                "to_user_id": to_user_id,
                "client_id": f"wcfLink-{secrets.randbits(63)}",
                "message_type": 2,
                "message_state": 2,
                "item_list": [{"type": 1, "text_item": {"text": text}}],
                "context_token": context_token,
            },
            "base_info": {"channel_version": self.channel_version},
        }
        out = self._post_json(f"{base_url.rstrip('/')}/ilink/bot/sendmessage", token, payload)
        self._raise_if_send_failed(out)

    def get_upload_url(self, base_url: str, token: str, payload: dict[str, Any]) -> dict[str, Any]:
        payload = dict(payload)
        payload["base_info"] = {"channel_version": self.channel_version}
        return self._post_json(f"{base_url.rstrip('/')}/ilink/bot/getuploadurl", token, payload)

    def upload_local_media(
        self,
        cdn_base_url: str,
        base_url: str,
        token: str,
        to_user_id: str,
        file_path: str,
        media_type: int,
    ) -> dict[str, Any]:
        plaintext = Path(file_path).read_bytes()
        raw_md5 = hashlib.md5(plaintext).hexdigest()
        aes_key = secrets.token_bytes(16)
        file_key = secrets.token_hex(16)
        ciphertext = encrypt_aes_ecb(plaintext, aes_key)
        upload_resp = self.get_upload_url(
            base_url,
            token,
            {
                "filekey": file_key,
                "media_type": media_type,
                "to_user_id": to_user_id,
                "rawsize": len(plaintext),
                "rawfilemd5": raw_md5,
                "filesize": len(ciphertext),
                "no_need_thumb": True,
                "aeskey": aes_key.hex(),
            },
        )
        upload_param = str(upload_resp.get("upload_param", ""))
        if not upload_param:
            raise RuntimeError("getuploadurl returned empty upload_param")
        download_param = self._upload_ciphertext_to_cdn(cdn_base_url, upload_param, file_key, ciphertext)
        return {
            "download_encrypted_query_param": download_param,
            "aes_key_hex": aes_key.hex(),
            "plain_size": len(plaintext),
            "cipher_size": len(ciphertext),
        }

    def send_image_message(
        self, base_url: str, token: str, to_user_id: str, context_token: str, text: str, uploaded: dict[str, Any]
    ) -> None:
        self._send_media_items(
            base_url,
            token,
            to_user_id,
            context_token,
            text,
            {
                "type": 2,
                "image_item": {
                    "media": {
                        "encrypt_query_param": uploaded["download_encrypted_query_param"],
                        "aes_key": base64.b64encode(uploaded["aes_key_hex"].encode("utf-8")).decode("ascii"),
                        "encrypt_type": 1,
                    },
                    "mid_size": uploaded["cipher_size"],
                },
            },
        )

    def send_video_message(
        self, base_url: str, token: str, to_user_id: str, context_token: str, text: str, uploaded: dict[str, Any]
    ) -> None:
        self._send_media_items(
            base_url,
            token,
            to_user_id,
            context_token,
            text,
            {
                "type": 5,
                "video_item": {
                    "media": {
                        "encrypt_query_param": uploaded["download_encrypted_query_param"],
                        "aes_key": base64.b64encode(uploaded["aes_key_hex"].encode("utf-8")).decode("ascii"),
                        "encrypt_type": 1,
                    },
                    "video_size": uploaded["cipher_size"],
                },
            },
        )

    def send_file_message(
        self,
        base_url: str,
        token: str,
        to_user_id: str,
        context_token: str,
        text: str,
        file_name: str,
        uploaded: dict[str, Any],
    ) -> None:
        self._send_media_items(
            base_url,
            token,
            to_user_id,
            context_token,
            text,
            {
                "type": 4,
                "file_item": {
                    "media": {
                        "encrypt_query_param": uploaded["download_encrypted_query_param"],
                        "aes_key": base64.b64encode(uploaded["aes_key_hex"].encode("utf-8")).decode("ascii"),
                        "encrypt_type": 1,
                    },
                    "file_name": file_name,
                    "len": str(uploaded["plain_size"]),
                },
            },
        )

    def send_voice_message(
        self,
        base_url: str,
        token: str,
        to_user_id: str,
        context_token: str,
        text: str,
        encode_type: int,
        uploaded: dict[str, Any],
    ) -> None:
        self._send_media_items(
            base_url,
            token,
            to_user_id,
            context_token,
            text,
            {
                "type": 3,
                "voice_item": {
                    "media": {
                        "encrypt_query_param": uploaded["download_encrypted_query_param"],
                        "aes_key": base64.b64encode(uploaded["aes_key_hex"].encode("utf-8")).decode("ascii"),
                        "encrypt_type": 1,
                    },
                    "encode_type": encode_type,
                    "text": "",
                },
            },
        )

    def download_message_media(self, cdn_base_url: str, item: dict[str, Any]) -> tuple[bytes, str, str]:
        item_type = int(item.get("type", 0))
        if item_type == 2:
            image_item = item.get("image_item") or {}
            media = image_item.get("media") or {}
            aes_key = media.get("aes_key", "")
            if image_item.get("aeskey"):
                aes_key = base64.b64encode(str(image_item["aeskey"]).encode("utf-8")).decode("ascii")
            buf = self._download_cdn_media(cdn_base_url, str(media.get("encrypt_query_param", "")), aes_key)
            mime = detect_mime(buf, ".jpg")
            return buf, "image" + extension_from_mime(mime, ".jpg"), mime
        if item_type == 3:
            voice_item = item.get("voice_item") or {}
            media = voice_item.get("media") or {}
            buf = self._download_cdn_media(cdn_base_url, str(media.get("encrypt_query_param", "")), str(media.get("aes_key", "")))
            return buf, "voice.silk", "audio/silk"
        if item_type == 4:
            file_item = item.get("file_item") or {}
            media = file_item.get("media") or {}
            file_name = str(file_item.get("file_name") or "file.bin")
            buf = self._download_cdn_media(cdn_base_url, str(media.get("encrypt_query_param", "")), str(media.get("aes_key", "")))
            return buf, file_name, detect_mime(buf, Path(file_name).suffix)
        if item_type == 5:
            video_item = item.get("video_item") or {}
            media = video_item.get("media") or {}
            buf = self._download_cdn_media(cdn_base_url, str(media.get("encrypt_query_param", "")), str(media.get("aes_key", "")))
            return buf, "video.mp4", "video/mp4"
        raise RuntimeError(f"unsupported media item type {item_type}")

    def _send_media_items(
        self,
        base_url: str,
        token: str,
        to_user_id: str,
        context_token: str,
        text: str,
        media_item: dict[str, Any],
    ) -> None:
        items: list[dict[str, Any]] = []
        if text.strip():
            items.append({"type": 1, "text_item": {"text": text}})
        items.append(media_item)
        for item in items:
            payload = {
                "msg": {
                    "from_user_id": "",
                    "to_user_id": to_user_id,
                    "client_id": f"wcfLink-{secrets.randbits(63)}",
                    "message_type": 2,
                    "message_state": 2,
                    "item_list": [item],
                    "context_token": context_token,
                },
                "base_info": {"channel_version": self.channel_version},
            }
            out = self._post_json(f"{base_url.rstrip('/')}/ilink/bot/sendmessage", token, payload)
            self._raise_if_send_failed(out)

    def _post_json(self, endpoint: str, token: str, body: dict[str, Any]) -> dict[str, Any]:
        payload = json.dumps(body).encode("utf-8")
        req = request.Request(endpoint, data=payload, method="POST")
        return self._do_json(req, token=token, payload=payload)

    def _do_json(self, req: request.Request, token: str, payload: bytes | None) -> dict[str, Any]:
        req.add_header("Content-Type", "application/json")
        req.add_header("AuthorizationType", "ilink_bot_token")
        req.add_header("X-WECHAT-UIN", random_wechat_uin())
        if payload is not None:
            req.add_header("Content-Length", str(len(payload)))
        if token.strip():
            req.add_header("Authorization", f"Bearer {token.strip()}")
        try:
            with request.urlopen(req, timeout=self.timeout) as resp:
                raw = resp.read()
        except error.HTTPError as exc:
            raw = exc.read()
            raise RuntimeError(f"ilink http {exc.code}: {raw.decode('utf-8', errors='replace')}") from exc
        except error.URLError as exc:
            raise RuntimeError(str(exc.reason)) from exc
        if not raw:
            return {}
        return json.loads(raw.decode("utf-8"))

    def _upload_ciphertext_to_cdn(self, cdn_base_url: str, upload_param: str, file_key: str, ciphertext: bytes) -> str:
        endpoint = (
            f"{cdn_base_url.rstrip('/')}/upload?encrypted_query_param={parse.quote(upload_param)}&filekey={parse.quote(file_key)}"
        )
        req = request.Request(endpoint, data=ciphertext, method="POST", headers={"Content-Type": "application/octet-stream"})
        try:
            with request.urlopen(req, timeout=self.timeout) as resp:
                resp.read()
                download_param = resp.headers.get("x-encrypted-param", "")
        except error.HTTPError as exc:
            raise RuntimeError(f"cdn upload http {exc.code}: {exc.read().decode('utf-8', errors='replace')}") from exc
        if not download_param.strip():
            raise RuntimeError("cdn upload response missing x-encrypted-param")
        return download_param

    def _download_cdn_media(self, cdn_base_url: str, encrypted_query_param: str, aes_key_base64: str) -> bytes:
        endpoint = f"{cdn_base_url.rstrip('/')}/download?encrypted_query_param={parse.quote(encrypted_query_param)}"
        req = request.Request(endpoint, method="GET")
        try:
            with request.urlopen(req, timeout=self.timeout) as resp:
                raw = resp.read()
        except error.HTTPError as exc:
            raise RuntimeError(f"cdn download http {exc.code}: {exc.read().decode('utf-8', errors='replace')}") from exc
        if not aes_key_base64.strip():
            return raw
        return decrypt_aes_ecb(raw, parse_aes_key(aes_key_base64))

    @staticmethod
    def _raise_if_send_failed(payload: dict[str, Any]) -> None:
        ret = int(payload.get("ret", 0) or 0)
        err_code = int(payload.get("errcode", 0) or 0)
        if ret == 0 and err_code == 0:
            return
        err_msg = str(payload.get("errmsg") or "sendmessage returned non-zero status")
        raise RuntimeError(f"{err_msg} (ret={ret} errcode={err_code})")


def random_wechat_uin() -> str:
    return base64.b64encode(os.urandom(8)).decode("ascii")


def encrypt_aes_ecb(plaintext: bytes, key: bytes) -> bytes:
    cipher = _new_aes_cipher(key)
    block_size = 16
    padding = block_size - len(plaintext) % block_size
    padded = plaintext + bytes([padding]) * padding
    return b"".join(cipher.encrypt(padded[i : i + block_size]) for i in range(0, len(padded), block_size))


def decrypt_aes_ecb(ciphertext: bytes, key: bytes) -> bytes:
    if len(ciphertext) % 16 != 0:
        raise RuntimeError(f"ciphertext size {len(ciphertext)} is not a multiple of block size")
    cipher = _new_aes_cipher(key)
    raw = b"".join(cipher.decrypt(ciphertext[i : i + 16]) for i in range(0, len(ciphertext), 16))
    padding = raw[-1]
    if padding <= 0 or padding > 16 or raw[-padding:] != bytes([padding]) * padding:
        raise RuntimeError("invalid PKCS7 padding")
    return raw[:-padding]


def parse_aes_key(aes_key_base64: str) -> bytes:
    decoded = base64.b64decode(aes_key_base64)
    if len(decoded) == 16:
        return decoded
    if len(decoded) == 32 and all(chr(b) in "0123456789abcdefABCDEF" for b in decoded):
        return bytes.fromhex(decoded.decode("ascii"))
    raise RuntimeError(f"unexpected aes_key length {len(decoded)}")


def _new_aes_cipher(key: bytes) -> Any:
    try:
        from Crypto.Cipher import AES
    except ImportError as exc:
        raise RuntimeError("media encryption requires installing the 'pycryptodome' dependency") from exc
    return AES.new(key, AES.MODE_ECB)


def detect_mime(buf: bytes, fallback_ext: str) -> str:
    guessed, _ = mimetypes.guess_type(f"dummy{fallback_ext}")
    if guessed:
        return guessed
    return "application/octet-stream"


def extension_from_mime(mime: str, fallback: str) -> str:
    return mimetypes.guess_extension(mime) or fallback



from __future__ import annotations

from importlib import metadata

from .models import VersionInfo


def current() -> VersionInfo:
    try:
        version = metadata.version("wcflink")
    except metadata.PackageNotFoundError:
        version = "0.1.0"
    return VersionInfo(version=version, commit=None, build_time=None, modified=False)



from __future__ import annotations

import json
import logging
import mimetypes
from datetime import datetime
from pathlib import Path
import time
from typing import Any
from urllib import request

from .config import Config, save_file_settings
from .ilink_client import ILinkClient
from .models import Account, LoginSession, Settings, utc_now_iso
from .poller import PollerManager
from .store import Store, detect_event_type, extract_body_text


class RuntimeState:
    def __init__(self, cfg: Config) -> None:
        self._settings = Settings(listen_addr=cfg.listen_addr, webhook_url=cfg.webhook_url)

    def settings(self) -> Settings:
        return Settings.from_dict(self._settings.to_dict())

    def update_settings(self, settings: Settings) -> None:
        self._settings = Settings.from_dict(settings.to_dict())


class Service:
    def __init__(
        self,
        cfg: Config,
        logger: logging.Logger,
        store: Store,
        ilink_client: ILinkClient,
        runtime: RuntimeState,
        pollers: PollerManager | None = None,
    ) -> None:
        self.cfg = cfg
        self.logger = logger
        self.store = store
        self.ilink_client = ilink_client
        self.runtime = runtime
        self.pollers = pollers

    def start_login(self, base_url: str = "") -> LoginSession:
        if not base_url:
            base_url = self.cfg.default_base_url
        qr = self.ilink_client.fetch_qrcode(base_url)
        now = utc_now_iso()
        session = LoginSession(
            session_id=f"login_{time.time_ns()}",
            base_url=base_url,
            qr_code=str(qr.get("qrcode") or ""),
            qr_code_url=str(qr.get("qrcode_img_content") or ""),
            status="wait",
            started_at=now,
            updated_at=now,
        )
        self.store.create_login_session(session)
        return session

    def get_login_status(self, session_id: str) -> LoginSession:
        session = self.store.get_login_session(session_id)
        if session.status == "confirmed":
            return session
        status = self.ilink_client.fetch_qrcode_status(session.base_url, session.qr_code)
        if not str(status.get("status") or ""):
            return session
        if str(status.get("status")) == "confirmed":
            self.store.complete_login_session(session_id, status)
            if self.pollers is not None:
                try:
                    self.pollers.start_account(self.store.get_account(str(status.get("ilink_bot_id") or "")))
                except LookupError:
                    pass
            return self.store.get_login_session(session_id)
        self.store.update_login_session_status(session_id, str(status.get("status") or session.status), "")
        return self.store.get_login_session(session_id)

    def get_login_session(self, session_id: str) -> LoginSession:
        return self.store.get_login_session(session_id)

    def list_accounts(self) -> list[Account]:
        return self.store.list_accounts()

    def list_events(self, after_id: int = 0, limit: int = 100):
        return self.store.list_events(after_id, limit)

    def list_logs(self, after_id: int = 0, limit: int = 100):
        return self.store.list_logs(after_id, limit)

    def get_settings(self) -> Settings:
        return self.runtime.settings()

    def update_settings(self, settings: Settings) -> Settings:
        save_file_settings(self.cfg.settings_path, settings.listen_addr, settings.webhook_url)
        self.runtime.update_settings(settings)
        self.store.add_log("INFO", "settings updated", "settings", "")
        return self.runtime.settings()

    def logout_account(self, account_id: str) -> None:
        if self.pollers is not None:
            self.pollers.stop_account(account_id)
        self.store.delete_account(account_id)
        self.store.add_log("INFO", "account disconnected locally", "account", json.dumps({"account_id": account_id}))

    def send_text(self, account_id: str, to_user_id: str, text: str, context_token: str = "") -> None:
        account = self.store.get_account(account_id)
        context_token = self._resolve_context_token(account_id, to_user_id, context_token)
        if not context_token.strip():
            raise RuntimeError("context token not found for this user; current text sending only supports replying to users who have already sent a message")
        self.ilink_client.send_text_message(account.base_url, account.token, to_user_id, text, context_token)
        raw = json.dumps(
            {"to_user_id": to_user_id, "text": text, "context_token": context_token},
            ensure_ascii=False,
        )
        self.store.create_outbound_event(account_id, "text", to_user_id, context_token, text, "", "", "", raw)
        self.store.add_log("INFO", "outbound text sent", "message", raw)

    def send_media(
        self,
        account_id: str,
        to_user_id: str,
        media_type: str,
        file_path: str,
        text: str = "",
        context_token: str = "",
    ) -> None:
        account = self.store.get_account(account_id)
        context_token = self._resolve_context_token(account_id, to_user_id, context_token)
        if not context_token.strip():
            raise RuntimeError("context token not found for this user; media sending only supports replying to users who have already sent a message")

        normalized_type, upload_type = normalize_media_send_type(media_type, file_path)
        uploaded = self.ilink_client.upload_local_media(
            self.cfg.cdn_base_url,
            account.base_url,
            account.token,
            to_user_id,
            file_path,
            upload_type,
        )
        file_name = Path(file_path).name
        if normalized_type == "image":
            self.ilink_client.send_image_message(account.base_url, account.token, to_user_id, context_token, text, uploaded)
        elif normalized_type == "video":
            self.ilink_client.send_video_message(account.base_url, account.token, to_user_id, context_token, text, uploaded)
        elif normalized_type == "file":
            self.ilink_client.send_file_message(
                account.base_url, account.token, to_user_id, context_token, text, file_name, uploaded
            )
        elif normalized_type == "voice":
            self.ilink_client.send_voice_message(
                account.base_url,
                account.token,
                to_user_id,
                context_token,
                text,
                detect_voice_encode_type(file_path),
                uploaded,
            )
        else:
            raise RuntimeError(f"unsupported media type {normalized_type!r}")

        mime_type = detect_outbound_mime(normalized_type, file_path)
        raw = json.dumps(
            {
                "to_user_id": to_user_id,
                "file_path": file_path,
                "media_type": normalized_type,
                "text": text,
                "context_token": context_token,
            },
            ensure_ascii=False,
        )
        self.store.create_outbound_event(
            account_id, normalized_type, to_user_id, context_token, text, file_path, file_name, mime_type, raw
        )
        self.store.add_log("INFO", "outbound media sent", "message", raw)

    def handle_inbound_message(self, account: Account, message: dict[str, Any]) -> None:
        media_path = ""
        media_file_name = ""
        media_mime_type = ""
        media_item = first_inbound_media_item(message)
        if media_item is not None:
            try:
                media_bytes, suggested_file_name, mime_type = self.ilink_client.download_message_media(
                    self.cfg.cdn_base_url, media_item
                )
                media_path, media_file_name, media_mime_type = self._save_inbound_media(
                    account.account_id,
                    int(message.get("message_id", 0) or 0),
                    str(message.get("from_user_id") or ""),
                    suggested_file_name,
                    mime_type,
                    media_bytes,
                )
            except Exception as exc:
                self.logger.warning("download inbound media failed: %s", exc)
                self.store.add_log(
                    "ERROR",
                    "download inbound media failed",
                    "media",
                    json.dumps(
                        {
                            "account_id": account.account_id,
                            "message_id": int(message.get("message_id", 0) or 0),
                            "err": str(exc),
                        },
                        ensure_ascii=False,
                    ),
                )

        self.store.save_inbound_message(account.account_id, message, media_path, media_file_name, media_mime_type)

        payload = json.dumps(
            {
                "account_id": account.account_id,
                "base_url": account.base_url,
                "event_type": detect_event_type(message),
                "body_text": extract_body_text(message),
                "from_user_id": str(message.get("from_user_id") or ""),
                "to_user_id": str(message.get("to_user_id") or ""),
                "message_id": int(message.get("message_id", 0) or 0),
                "context_token": str(message.get("context_token") or ""),
                "media_path": media_path,
                "media_file_name": media_file_name,
                "media_mime_type": media_mime_type,
                "raw_message": message,
                "received_at": utc_now_iso(),
            },
            ensure_ascii=False,
        ).encode("utf-8")

        settings = self.runtime.settings()
        if not settings.webhook_url:
            self.store.add_log("INFO", "inbound message received", "inbound", payload.decode("utf-8"))
            return
        self._deliver_webhook(settings.webhook_url, payload)

    def _resolve_context_token(self, account_id: str, to_user_id: str, context_token: str) -> str:
        if context_token.strip():
            return context_token
        try:
            return self.store.get_peer_context(account_id, to_user_id).context_token
        except LookupError:
            return ""

    def _save_inbound_media(
        self,
        account_id: str,
        message_id: int,
        from_user_id: str,
        file_name: str,
        mime_type: str,
        data: bytes,
    ) -> tuple[str, str, str]:
        media_root = Path(self.cfg.media_dir)
        now = datetime.utcnow()
        target_dir = media_root / sanitize_path_segment(account_id) / now.strftime("%Y") / now.strftime("%m") / now.strftime("%d")
        target_dir.mkdir(parents=True, exist_ok=True)

        safe_name = sanitize_file_name(file_name) or "media"
        suffix = Path(safe_name).suffix or extension_for_mime(mime_type) or ".bin"
        base = Path(safe_name).stem
        prefix = str(message_id or __import__("time").time_ns())
        if from_user_id:
            prefix = f"{prefix}_{sanitize_path_segment(from_user_id)}"
        final_name = f"{prefix}_{base}{suffix}"
        full_path = target_dir / final_name
        full_path.write_bytes(data)
        return str(full_path), final_name, mime_type

    def _deliver_webhook(self, webhook_url: str, payload: bytes) -> None:
        req = request.Request(webhook_url, data=payload, method="POST", headers={"Content-Type": "application/json"})
        try:
            with request.urlopen(req, timeout=10) as resp:
                if resp.status < 200 or resp.status >= 300:
                    raise RuntimeError(f"webhook delivery failed with status {resp.status}")
        except Exception as exc:
            self.store.add_log("ERROR", "webhook delivery failed", "webhook", str(exc))
            return
        self.store.add_log("INFO", "webhook delivered", "webhook", "")


def normalize_media_send_type(media_type: str, file_path: str) -> tuple[str, int]:
    value = media_type.strip().lower()
    if not value:
        suffix = Path(file_path).suffix.lower()
        if suffix in {".jpg", ".jpeg", ".png", ".gif", ".webp"}:
            value = "image"
        elif suffix in {".mp4", ".mov", ".m4v"}:
            value = "video"
        elif suffix in {".silk", ".amr", ".mp3", ".ogg", ".wav", ".m4a"}:
            value = "voice"
        else:
            value = "file"
    mapping = {"image": 1, "video": 2, "file": 3, "voice": 4}
    if value not in mapping:
        raise RuntimeError(f"unsupported media type {media_type!r}")
    return value, mapping[value]


def first_inbound_media_item(message: dict[str, Any]) -> dict[str, Any] | None:
    for item in list(message.get("item_list") or []):
        entry = dict(item)
        if int(entry.get("type", 0) or 0) in {2, 3, 4, 5}:
            return entry
    return None


def detect_outbound_mime(media_type: str, file_path: str) -> str:
    if media_type == "image":
        suffix = Path(file_path).suffix.lower()
        if suffix == ".png":
            return "image/png"
        if suffix == ".gif":
            return "image/gif"
        return "image/jpeg"
    if media_type == "video":
        return "video/mp4"
    if media_type == "voice":
        suffix = Path(file_path).suffix.lower()
        return {
            ".amr": "audio/amr",
            ".mp3": "audio/mpeg",
            ".ogg": "audio/ogg",
        }.get(suffix, "audio/silk")
    return "application/octet-stream"


def sanitize_path_segment(value: str) -> str:
    out = str(value).strip()
    for char in '/\\:*?"<>|@':
        out = out.replace(char, "_")
    return out or "unknown"


def sanitize_file_name(value: str) -> str:
    name = Path(str(value).strip()).name
    if name in {"", ".", "/"}:
        return ""
    return sanitize_path_segment(name)


def extension_for_mime(mime_type: str) -> str:
    return mimetypes.guess_extension(mime_type) or ""


def detect_voice_encode_type(file_path: str) -> int:
    return {
        ".amr": 5,
        ".mp3": 7,
        ".ogg": 8,
    }.get(Path(file_path).suffix.lower(), 6)



from .config import Config, load as load_config
from .client import WcfLinkClient
from .engine import Engine, current_version
from .exceptions import WcfLinkAPIError, WcfLinkClientError
from .models import (
    Account,
    Event,
    LoginSession,
    LogEntry,
    Settings,
    VersionInfo,
)

__all__ = [
    "Config",
    "current_version",
    "Engine",
    "load_config",
    "WcfLinkAPIError",
    "WcfLinkClient",
    "WcfLinkClientError",
    "Account",
    "Event",
    "LoginSession",
    "LogEntry",
    "Settings",
    "VersionInfo",
]



from __future__ import annotations

import logging
import time

from .config import Config, load
from .ilink_client import ILinkClient
from .models import Account, Event, LogEntry, LoginSession, Settings, VersionInfo
from .poller import PollerManager
from .server import APIServer
from .service import RuntimeState, Service
from .store import Store
from .version import current


class Engine:
    def __init__(self, cfg: Config | None = None, logger: logging.Logger | None = None) -> None:
        self.cfg = cfg or load()
        self.logger = logger or logging.getLogger("wcflink")
        self.store = Store(self.cfg.db_path)
        self.ilink_client = ILinkClient(self.cfg.channel_version, self.cfg.poll_timeout + 10.0)
        self.runtime = RuntimeState(self.cfg)
        self.pollers = PollerManager(self.store, self.ilink_client, self.logger)
        self.service = Service(self.cfg, self.logger, self.store, self.ilink_client, self.runtime, self.pollers)
        self.pollers.on_message = self.service.handle_inbound_message
        self.server = APIServer(self.service, self.logger, self.cfg.listen_addr)
        self._started = False

    def start_background(self) -> None:
        if self._started:
            return
        self.store.ping()
        self.pollers.start_enabled_accounts()
        self.server.start()
        self._started = True

    def run(self) -> None:
        self.start_background()
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            self.logger.info("shutdown requested")
        finally:
            self.shutdown()

    def shutdown(self) -> None:
        self.pollers.stop_all()
        self.server.shutdown()
        self.store.close()

    def start_login(self, base_url: str = "") -> LoginSession:
        return self.service.start_login(base_url)

    def get_login_status(self, session_id: str) -> LoginSession:
        return self.service.get_login_status(session_id)

    def get_login_session(self, session_id: str) -> LoginSession:
        return self.service.get_login_session(session_id)

    def list_accounts(self) -> list[Account]:
        return self.service.list_accounts()

    def list_events(self, after_id: int = 0, limit: int = 100) -> list[Event]:
        return self.service.list_events(after_id, limit)

    def list_logs(self, after_id: int = 0, limit: int = 100) -> list[LogEntry]:
        return self.service.list_logs(after_id, limit)

    def get_settings(self) -> Settings:
        return self.service.get_settings()

    def update_settings(self, settings: Settings) -> Settings:
        return self.service.update_settings(settings)

    def send_text(self, account_id: str, to_user_id: str, text: str, context_token: str = "") -> None:
        self.service.send_text(account_id, to_user_id, text, context_token)

    def send_media(
        self,
        account_id: str,
        to_user_id: str,
        media_type: str,
        file_path: str,
        text: str = "",
        context_token: str = "",
    ) -> None:
        self.service.send_media(account_id, to_user_id, media_type, file_path, text, context_token)

    def logout_account(self, account_id: str) -> None:
        self.service.logout_account(account_id)


def load_config() -> Config:
    return load()


def current_version() -> VersionInfo:
    return current()



from __future__ import annotations

import argparse
import json
import logging
import sys
from pathlib import Path

from .client import WcfLinkClient
from .config import load
from .engine import Engine, current_version
from .exceptions import WcfLinkAPIError


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="wcflink", description="Python implementation of wcfLink")
    parser.add_argument("--base-url", default="http://127.0.0.1:17890", help="wcfLink service base URL")

    sub = parser.add_subparsers(dest="command", required=True)

    serve = sub.add_parser("serve")
    serve.add_argument("--listen-addr", default=None)
    serve.add_argument("--state-dir", default=None)
    serve.add_argument("--media-dir", default=None)
    serve.add_argument("--db-path", default=None)
    serve.add_argument("--upstream-base-url", default=None)
    serve.add_argument("--cdn-base-url", default=None)
    serve.add_argument("--channel-version", default=None)
    serve.add_argument("--poll-timeout", type=float, default=None)
    serve.add_argument("--webhook-url", default=None)
    serve.add_argument("--log-level", default=None)

    sub.add_parser("version")
    sub.add_parser("api-version")
    sub.add_parser("accounts")

    events = sub.add_parser("events")
    events.add_argument("--after-id", type=int, default=0)
    events.add_argument("--limit", type=int, default=100)

    logs = sub.add_parser("logs")
    logs.add_argument("--after-id", type=int, default=0)
    logs.add_argument("--limit", type=int, default=100)

    login = sub.add_parser("login")
    login_sub = login.add_subparsers(dest="login_command", required=True)

    login_start = login_sub.add_parser("start")
    login_start.add_argument("--upstream-base-url", default="", help="Override iLink upstream base URL")

    login_status = login_sub.add_parser("status")
    login_status.add_argument("session_id")

    login_qr = login_sub.add_parser("qr")
    login_qr.add_argument("session_id")
    login_qr.add_argument("-o", "--output", required=True, help="PNG output path")

    send_text = sub.add_parser("send-text")
    send_text.add_argument("--account-id", required=True)
    send_text.add_argument("--to-user-id", required=True)
    send_text.add_argument("--text", required=True)
    send_text.add_argument("--context-token", default="")

    send_media = sub.add_parser("send-media")
    send_media.add_argument("--account-id", required=True)
    send_media.add_argument("--to-user-id", required=True)
    send_media.add_argument("--file-path", required=True)
    send_media.add_argument("--type", default="")
    send_media.add_argument("--text", default="")
    send_media.add_argument("--context-token", default="")

    return parser


def main() -> int:
    parser = build_parser()
    args = parser.parse_args()

    if args.command == "serve":
        return run_server(args)

    try:
        match args.command:
            case "version":
                print_json(current_version().to_dict())
            case "api-version":
                client = WcfLinkClient(args.base_url)
                print_json(client.version().to_dict())
            case "accounts":
                client = WcfLinkClient(args.base_url)
                print_json([item.to_dict() for item in client.list_accounts()])
            case "events":
                client = WcfLinkClient(args.base_url)
                print_json([item.to_dict() for item in client.list_events(args.after_id, args.limit)])
            case "logs":
                client = WcfLinkClient(args.base_url)
                print_json(client.list_logs(args.after_id, args.limit))
            case "login":
                client = WcfLinkClient(args.base_url)
                return handle_login(args, client)
            case "send-text":
                client = WcfLinkClient(args.base_url)
                print_json(
                    client.send_text(
                        account_id=args.account_id,
                        to_user_id=args.to_user_id,
                        text=args.text,
                        context_token=args.context_token,
                    )
                )
            case "send-media":
                client = WcfLinkClient(args.base_url)
                print_json(
                    client.send_media(
                        account_id=args.account_id,
                        to_user_id=args.to_user_id,
                        file_path=args.file_path,
                        media_type=args.type,
                        text=args.text,
                        context_token=args.context_token,
                    )
                )
            case _:
                parser.error("unsupported command")
    except WcfLinkAPIError as exc:
        sys.stderr.write(f"wcfLink API error: {exc}\n")
        return 1

    return 0


def handle_login(args: argparse.Namespace, client: WcfLinkClient) -> int:
    match args.login_command:
        case "start":
            print_json(client.start_login(base_url=args.upstream_base_url).to_dict())
        case "status":
            print_json(client.get_login_status(args.session_id).to_dict())
        case "qr":
            output = Path(args.output)
            output.write_bytes(client.get_login_qr(args.session_id))
            print(str(output))
        case _:
            raise ValueError("unsupported login command")
    return 0


def print_json(value: object) -> None:
    print(json.dumps(value, ensure_ascii=False, indent=2))


def run_server(args: argparse.Namespace) -> int:
    cfg = load().with_overrides(
        listen_addr=args.listen_addr,
        state_dir=args.state_dir,
        media_dir=args.media_dir,
        db_path=args.db_path,
        default_base_url=args.upstream_base_url,
        cdn_base_url=args.cdn_base_url,
        channel_version=args.channel_version,
        poll_timeout=args.poll_timeout,
        webhook_url=args.webhook_url,
        log_level=args.log_level,
    )
    if args.state_dir is not None:
        if args.media_dir is None:
            cfg = cfg.with_overrides(media_dir=str(Path(args.state_dir) / "media"))
        if args.db_path is None:
            cfg = cfg.with_overrides(db_path=str(Path(args.state_dir) / "wcflink.db"))
        cfg = cfg.with_overrides(settings_path=str(Path(args.state_dir) / "settings.json"))

    logging.basicConfig(
        level=getattr(logging, str(cfg.log_level).upper(), logging.INFO),
        format="%(asctime)s %(levelname)s %(name)s %(message)s",
    )
    logger = logging.getLogger("wcflink")
    logger.info("starting wcflink on %s", cfg.listen_addr)
    logger.info("version %s", current_version().version)
    engine = Engine(cfg=cfg, logger=logger)
    engine.run()
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



class WcfLinkClientError(Exception):
    """Base exception for the Python wcfLink client."""


class WcfLinkAPIError(WcfLinkClientError):
    """Raised when the wcfLink HTTP API returns a non-2xx response."""

    def __init__(self, status_code: int, message: str, body: object | None = None) -> None:
        super().__init__(message)
        self.status_code = status_code
        self.body = body



from __future__ import annotations

import json
import os
from dataclasses import dataclass, replace
from pathlib import Path


DEFAULT_LISTEN_ADDR = "127.0.0.1:17890"
DEFAULT_BASE_URL = "https://ilinkai.weixin.qq.com"
DEFAULT_CDN_BASE_URL = "https://novac2c.cdn.weixin.qq.com/c2c"
DEFAULT_CHANNEL_VERSION = "2.0.1"
DEFAULT_POLL_TIMEOUT = 35.0


@dataclass(slots=True)
class Config:
    listen_addr: str
    state_dir: str
    media_dir: str
    db_path: str
    settings_path: str
    default_base_url: str
    cdn_base_url: str
    channel_version: str
    poll_timeout: float
    log_level: str
    open_browser: bool
    webhook_url: str

    def with_overrides(self, **overrides: object) -> "Config":
        normalized = {key: value for key, value in overrides.items() if value is not None}
        return replace(self, **normalized)


def default_state_dir() -> str:
    return str(Path.cwd() / "data")


def load_file_settings(path: str) -> dict[str, str]:
    try:
        return json.loads(Path(path).read_text(encoding="utf-8"))
    except (FileNotFoundError, json.JSONDecodeError):
        return {}


def save_file_settings(path: str, listen_addr: str, webhook_url: str) -> None:
    target = Path(path)
    target.parent.mkdir(parents=True, exist_ok=True)
    target.write_text(
        json.dumps({"listen_addr": listen_addr, "webhook_url": webhook_url}, ensure_ascii=False, indent=2),
        encoding="utf-8",
    )


def load() -> Config:
    state_dir = os.getenv("WCFLINK_STATE_DIR", default_state_dir())
    media_dir = os.getenv("WCFLINK_MEDIA_DIR", str(Path(state_dir) / "media"))
    db_path = os.getenv("WCFLINK_DB_PATH", str(Path(state_dir) / "wcflink.db"))
    settings_path = str(Path(state_dir) / "settings.json")
    file_settings = load_file_settings(settings_path)

    return Config(
        listen_addr=os.getenv("WCFLINK_LISTEN_ADDR", file_settings.get("listen_addr", DEFAULT_LISTEN_ADDR)),
        state_dir=state_dir,
        media_dir=media_dir,
        db_path=db_path,
        settings_path=settings_path,
        default_base_url=os.getenv("WCFLINK_BASE_URL", DEFAULT_BASE_URL),
        cdn_base_url=os.getenv("WCFLINK_CDN_BASE_URL", DEFAULT_CDN_BASE_URL),
        channel_version=os.getenv("WCFLINK_CHANNEL_VERSION", DEFAULT_CHANNEL_VERSION),
        poll_timeout=float(os.getenv("WCFLINK_POLL_TIMEOUT", str(DEFAULT_POLL_TIMEOUT))),
        log_level=os.getenv("WCFLINK_LOG_LEVEL", "INFO"),
        open_browser=_env_bool("WCFLINK_OPEN_BROWSER", False),
        webhook_url=os.getenv("WCFLINK_WEBHOOK_URL", file_settings.get("webhook_url", "")),
    )


def _env_bool(key: str, default: bool) -> bool:
    value = os.getenv(key)
    if value is None:
        return default
    return value.strip().lower() in {"1", "true", "yes", "on"}



from __future__ import annotations


def generate_qrcode_png(content: str) -> bytes:
    try:
        import qrcode
    except ImportError as exc:
        raise RuntimeError("qrcode support requires installing the 'qrcode[pil]' dependency") from exc
    image = qrcode.make(content)
    from io import BytesIO

    buffer = BytesIO()
    image.save(buffer, format="PNG")
    return buffer.getvalue()



from __future__ import annotations

import json
from typing import Any
from urllib import error, parse, request

from .exceptions import WcfLinkAPIError
from .models import Account, Event, LoginSession, Settings, VersionInfo


class WcfLinkClient:
    def __init__(self, base_url: str = "http://127.0.0.1:17890", timeout: float = 30.0) -> None:
        self.base_url = base_url.rstrip("/")
        self.timeout = timeout

    def version(self) -> VersionInfo:
        return VersionInfo.from_dict(self._request_json("GET", "/api/version"))

    def health_live(self) -> dict[str, Any]:
        return self._request_json("GET", "/health/live")

    def health_ready(self) -> dict[str, Any]:
        return self._request_json("GET", "/health/ready")

    def start_login(self, base_url: str = "") -> LoginSession:
        payload = {"base_url": base_url} if base_url else {}
        return LoginSession.from_dict(self._request_json("POST", "/api/accounts/login/start", payload))

    def get_login_status(self, session_id: str) -> LoginSession:
        query = parse.urlencode({"session_id": session_id})
        return LoginSession.from_dict(self._request_json("GET", f"/api/accounts/login/status?{query}"))

    def get_login_qr(self, session_id: str) -> bytes:
        query = parse.urlencode({"session_id": session_id})
        return self._request_bytes("GET", f"/api/accounts/login/qr?{query}")

    def list_accounts(self) -> list[Account]:
        data = self._request_json("GET", "/api/accounts")
        return [Account.from_dict(item) for item in data.get("items", [])]

    def list_events(self, after_id: int = 0, limit: int = 100) -> list[Event]:
        query = parse.urlencode({"after_id": after_id, "limit": limit})
        data = self._request_json("GET", f"/api/events?{query}")
        return [Event.from_dict(item) for item in data.get("items", [])]

    def list_logs(self, after_id: int = 0, limit: int = 100) -> list[dict[str, Any]]:
        query = parse.urlencode({"after_id": after_id, "limit": limit})
        data = self._request_json("GET", f"/api/logs?{query}")
        return list(data.get("items", []))

    def get_settings(self) -> Settings:
        return Settings.from_dict(self._request_json("GET", "/api/settings"))

    def update_settings(self, listen_addr: str, webhook_url: str = "") -> dict[str, Any]:
        return self._request_json(
            "POST",
            "/api/settings",
            {"listen_addr": listen_addr, "webhook_url": webhook_url},
        )

    def send_text(
        self,
        account_id: str,
        to_user_id: str,
        text: str,
        context_token: str = "",
    ) -> dict[str, Any]:
        return self._request_json(
            "POST",
            "/api/messages/send-text",
            {
                "account_id": account_id,
                "to_user_id": to_user_id,
                "text": text,
                "context_token": context_token,
            },
        )

    def send_media(
        self,
        account_id: str,
        to_user_id: str,
        file_path: str,
        media_type: str = "",
        text: str = "",
        context_token: str = "",
    ) -> dict[str, Any]:
        return self._request_json(
            "POST",
            "/api/messages/send-media",
            {
                "account_id": account_id,
                "to_user_id": to_user_id,
                "file_path": file_path,
                "type": media_type,
                "text": text,
                "context_token": context_token,
            },
        )

    def _request_json(self, method: str, path: str, payload: dict[str, Any] | None = None) -> dict[str, Any]:
        raw = self._request_bytes(method, path, payload)
        if not raw:
            return {}
        return json.loads(raw.decode("utf-8"))

    def _request_bytes(self, method: str, path: str, payload: dict[str, Any] | None = None) -> bytes:
        data: bytes | None = None
        headers = {}
        if payload is not None:
            data = json.dumps(payload).encode("utf-8")
            headers["Content-Type"] = "application/json"

        req = request.Request(
            url=self.base_url + path,
            data=data,
            headers=headers,
            method=method,
        )
        try:
            with request.urlopen(req, timeout=self.timeout) as resp:
                return resp.read()
        except error.HTTPError as exc:
            body = exc.read()
            message = self._extract_error_message(body) or f"HTTP {exc.code}"
            raise WcfLinkAPIError(exc.code, message, self._parse_error_body(body)) from exc
        except error.URLError as exc:
            raise WcfLinkAPIError(0, str(exc.reason)) from exc

    @staticmethod
    def _parse_error_body(body: bytes) -> Any:
        if not body:
            return None
        try:
            return json.loads(body.decode("utf-8"))
        except Exception:
            return body.decode("utf-8", errors="replace")

    @classmethod
    def _extract_error_message(cls, body: bytes) -> str:
        parsed = cls._parse_error_body(body)
        if isinstance(parsed, dict):
            value = parsed.get("error")
            if value:
                return str(value)
        if isinstance(parsed, str):
            return parsed
        return ""
